import json
import random
import tempfile

from accession.backends import GCBackend


def filter_outputs_by_path(path, backend=None):
    bucket = path.split("gs://")[1].split("/")[0]
    if not backend:
        google_backend = GCBackend(bucket)
    else:
        google_backend = backend
    filtered = [
        file
        for file in list(google_backend.bucket.list_blobs())
        if path.split("gs://")[1] in file.id and ".json" in file.id
    ]

    for file in filtered:
        file.download_to_filename(file.public_url.split("/")[-1])

    return filtered


# Python equivalent of WDL's write_json
# for testing purposes
def write_json(json_data):
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    with open(temp_file.name, "a") as file:
        json.dump(json_data, file)
    return temp_file.name


def mutate_digits(md5sum_string):
    md5sum_list = list(md5sum_string)
    for i, character in enumerate(md5sum_list):
        print(i)
        print(character)
        if character.isdigit():
            if random.choice([True, False]):
                continue
            md5sum_list[i] = str((int(character) + 1) % 10)
        print(md5sum_list)
    return "".join(md5sum_list)


def string_to_number(string):
    if not isinstance(string, str):
        return string
    try:
        return int(string)
    except Exception:
        try:
            return float(string)
        except Exception:
            return string



"""
mutate_digits and write_json aren't actually used in the accession module, so they are
not yet tested here. filter_outputs_by_path needs refactoring to be readily testable.
"""
import os

from accession.helpers import filter_outputs_by_path, string_to_number

from .fixtures import MockBlob, MockBucket


def test_filter_outputs_by_path(mock_gc_backend, mocker):
    mocker.patch.object(mock_gc_backend.bucket, "list_blobs", mock_list_blobs)
    filtered = filter_outputs_by_path("gs://my-bucket/three/", backend=mock_gc_backend)
    filtered_paths = [file.path for file in filtered]
    assert filtered_paths == ["three/bazqux.json"]
    os.remove("bazqux.json")


def test_string_to_number_not_string_return_input():
    not_string = 3
    result = string_to_number(not_string)
    assert result == not_string


def test_string_to_number_stringy_int_returns_int():
    int_string = "3"
    result = string_to_number(int_string)
    assert isinstance(result, int)
    assert result == 3


def test_string_to_number_stringy_float_returns_float():
    float_string = "3.0"
    result = string_to_number(float_string)
    assert isinstance(result, float)
    assert result == 3.0


def test_string_to_number_non_number_string_returns_input():
    non_number_string = "3.0a"
    result = string_to_number(non_number_string)
    assert result == non_number_string


def mock_list_blobs():
    mock_bucket = MockBucket("my-bucket")
    for path in ["one/foobar.txt", "three/bazqux.json", "four/spam/eggs.json"]:
        yield MockBlob(path=path, bucket=mock_bucket)



import pytest


@pytest.mark.docker
@pytest.mark.filesystem
def test_accession_mirna_replicated(accessioner_factory):
    factory = accessioner_factory(
        metadata_file="mirna_replicated_metadata.json", assay_name="mirna"
    )
    accessioner, expected_files = next(factory)
    accessioner.accession_steps()
    validate_accessioning(
        accessioner, expected_files, expected_num_files=12, dataset="ENCSR715NEZ"
    )


@pytest.mark.docker
@pytest.mark.filesystem
def test_accession_mirna_unreplicated(accessioner_factory):
    factory = accessioner_factory(
        metadata_file="mirna_unreplicated_metadata.json", assay_name="mirna"
    )
    accessioner, expected_files = next(factory)
    accessioner.accession_steps()
    validate_accessioning(
        accessioner, expected_files, expected_num_files=6, dataset="ENCSR543MWW"
    )


def validate_accessioning(accessioner, expected_files, expected_num_files, dataset):
    all_files = accessioner.conn.get(f"/files/", database=True)["@graph"]
    accessioner.conn.get(f"/files/?dataset=/experiments/{dataset}/", database=True)[
        "@graph"
    ]
    all_quality_metrics = [
        accessioner.conn.get(qm["@id"], database=True, frame="object")
        for qm in accessioner.conn.get("/quality-metrics/", database=True)["@graph"]
    ]
    excluded_types = ("reads", "genome index", "genome reference")
    files = [
        f
        for f in all_files
        if f["output_type"] not in excluded_types
        and f["dataset"] == f"/experiments/{dataset}/"
    ]
    assert len(files) == expected_num_files
    shared_keys_to_skip = ["submitted_by", "date_created", "@id", "uuid"]
    file_keys_to_skip = [
        "title",
        "accession",
        "file_size",
        "cloud_metadata",
        "href",
        "s3_uri",
        "no_file_available",
        "analysis_step_version",
        "content_md5sum",
        "matching_md5sum",
        "audit",
        "schema_version",
    ]
    qm_keys_to_skip = ["step_run"]
    for keys_to_skip in (file_keys_to_skip, qm_keys_to_skip):
        keys_to_skip.extend(shared_keys_to_skip)
    for partial_file in files:
        file = accessioner.conn.get(
            partial_file["@id"], frame="embedded", database=True
        )
        quality_metrics = [
            qm for qm in all_quality_metrics if file["@id"] in qm["quality_metric_of"]
        ]
        aliases = file["aliases"]
        expected = [f for f in expected_files if f["aliases"] == aliases][0]
        for key, expected_value in expected.items():
            if key in file_keys_to_skip:
                continue
            elif key == "step_run":
                for step_run_key in ("analysis_step_version", "aliases"):
                    assert file[key][step_run_key] == expected_value[step_run_key]
                assert file[key]["status"] == "in progress"
            elif key == "status":
                assert file[key] == "uploading"
            elif key in ("award", "lab"):
                assert file[key]["@id"] == expected_value["@id"]
            elif key == "file_size":
                assert file[key] == 3
            elif key == "derived_from":
                derived_from_aliases = []
                for i in all_files:
                    if i["@id"] in file[key]:
                        db_file = accessioner.conn.get(i["@id"], database=True)
                        derived_from_aliases.append(
                            (db_file["aliases"], db_file["md5sum"])
                        )
                expected_aliases = [
                    (f["aliases"], f["md5sum"])
                    for f in expected_files
                    if f["@id"] in expected_value
                ]
                assert sorted(derived_from_aliases) == sorted(expected_aliases)
            elif key == "quality_metrics":
                if not expected_value:
                    assert not quality_metrics
                else:
                    for expected_qm in expected_value:
                        posted_qms = [
                            qm
                            for qm in quality_metrics
                            if qm["@type"] == expected_qm["@type"]
                        ]
                        assert len(posted_qms) == 1
                        posted_qm = posted_qms[0]
                        for qm_key, qm_value in expected_qm.items():
                            if qm_key == "quality_metric_of":
                                assert len(qm_value) == len(posted_qm[qm_key])
                            elif qm_key in qm_keys_to_skip:
                                continue
                            elif qm_key == "status":
                                assert posted_qm[qm_key] == "in progress"
                            else:
                                assert qm_value == posted_qm[qm_key]
            else:
                assert file[key] == expected_value



import argparse
import os
from typing import Optional, Tuple

from encode_utils.connection import Connection

from accession import __version__
from accession.accession import accession_factory
from accession.helpers import filter_outputs_by_path


def get_parser():
    parser = argparse.ArgumentParser(
        description="Accession pipeline outputs, \
                                                 download output metadata for scattering"
    )
    parser.add_argument(
        "--filter-from-path",
        type=str,
        default=None,
        help="path to a folder with pipeline run outputs",
    )
    parser.add_argument(
        "--accession-metadata",
        type=str,
        default=None,
        help="path to a metadata json output file",
    )
    parser.add_argument(
        "--pipeline-type",
        type=str,
        default=None,
        help=f"the type of pipeline run being accessioned, e.g. mirna or long_read_rna",
    )
    parser.add_argument(
        "--server", default="dev", help="Server the files will be accessioned to"
    )
    parser.add_argument("--lab", type=str, default=None, help="Lab")
    parser.add_argument("--award", type=str, default=None, help="Award")
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    return parser


def check_or_set_lab_award(lab: Optional[str], award: Optional[str]) -> Tuple[str, str]:
    error_msg = (
        "Could not identify {}, make sure you passed in the --{} command line argument"
        "or set the DCC_{} environment variable"
    )
    lab_prop = "lab"
    award_prop = "award"
    if not lab:
        try:
            lab = os.environ[f"DCC_{lab_prop.upper()}"]
        except KeyError as e:
            raise EnvironmentError(
                error_msg.format(lab_prop, lab_prop, lab_prop.upper())
            ) from e
    if not award:
        try:
            award = os.environ[f"DCC_{award_prop.upper()}"]
        except KeyError as e:
            raise EnvironmentError(
                error_msg.format(award_prop, award_prop, award_prop.upper())
            ) from e
    return lab, award


def main(args=None):
    parser = get_parser()
    args = parser.parse_args(args)
    lab, award = check_or_set_lab_award(args.lab, args.award)
    if args.filter_from_path:
        filter_outputs_by_path(args.filter_from_path)
        return
    connection = Connection(args.server)
    if all([args.pipeline_type, args.accession_metadata, lab, award, connection]):
        accessioner = accession_factory(
            args.pipeline_type, args.accession_metadata, connection, lab, award
        )
        accessioner.accession_steps()
        return
    print("Module called without proper arguments")


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest

from accession.accession import AccessionSteps


@pytest.mark.filesystem
def test_accession_steps_init(steps):
    """
    Once we access steps.content, the _steps is initialized.
    """
    assert not steps._steps
    assert steps.content
    assert (
        steps.content[0]["dcc_step_run"]
        == "/analysis-steps/long-read-rna-seq-alignments-step-v-1/"
    )
    assert steps._steps


@pytest.mark.filesystem
def test_accession_steps_path_to_json(steps):
    """
    The absolute path will vary across systems, so we only observe the last two pieces
    """
    assert steps.path_to_json.parts[-2:] == ("data", "long_rna_steps.json")


@pytest.mark.filesystem
def test_accession_steps_content(steps):
    assert isinstance(steps.content, list)
    assert all(isinstance(x, dict) for x in steps.content)


@pytest.fixture(scope="module")
def steps():
    current_dir = Path(__file__).resolve()
    steps_json_path = current_dir.parent / "data" / "long_rna_steps.json"
    return AccessionSteps(steps_json_path)



import pytest

from accession.quality_metric import QualityMetric


def test_quality_metric_no_file_id_raises(payload):
    with pytest.raises(Exception):
        QualityMetric(payload, file_id=None)


def test_quality_metric_init(payload):
    qm = QualityMetric(payload, "my_id")
    assert qm.files == ["my_id"]
    assert qm.payload == payload


@pytest.fixture
def payload():
    return {"foo": "bar"}



import json
import logging
import os
from abc import ABC, abstractmethod
from base64 import b64encode
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

from accession.analysis import Analysis, MetaData
from accession.file import GSFile
from accession.helpers import string_to_number
from accession.quality_metric import QualityMetric

ASSEMBLY = "assembly"
GENOME_ANNOTATION = "genome_annotation"


class AccessionSteps:
    def __init__(self, path_to_accession_step_json):
        self._path_to_accession_step_json = path_to_accession_step_json
        self._steps = None

    def _load_steps(self):
        if self._steps:
            return
        with open(self.path_to_json) as fp:
            self._steps = json.load(fp)

    @property
    def path_to_json(self):
        return self._path_to_accession_step_json

    @property
    def content(self):
        self._load_steps()
        return self._steps["accession.steps"]

    @property
    def raw_fastqs_keys(self) -> Optional[str]:
        self._load_steps()
        return self._steps.get("raw_fastqs_keys")


class Accession(ABC):
    """docstring for Accession
       Args:
        steps: AccessionSteps object
        analysis: Analysis object
        connection: Connection object
    """

    ACCESSION_LOG_KEY = "ACC_MSG"
    ASSEMBLIES = ["GRCh38", "mm10"]
    PROFILE_KEY = "_profile"

    def __init__(self, steps, analysis, connection, lab, award):
        self.analysis = analysis
        self.steps = steps
        self.backend = self.analysis.backend
        self.conn = connection
        self.COMMON_METADATA = {"lab": lab, "award": award}
        self._dataset = None
        self.new_files = []
        self.new_qcs = []
        self.raw_qcs = []
        self.logger = logging.getLogger(__name__)
        logging.basicConfig(
            filename="accession.log",
            format="%(asctime)s %(levelname)s %(message)s",
            level=logging.DEBUG,
        )

    @property
    @abstractmethod
    def QC_MAP(self):
        raise NotImplementedError("Derived classes should provide their own QC_MAPs")

    def get_step_run_id(self, encode_file):
        step_run = encode_file.get("step_run")
        if isinstance(step_run, str):
            step_run_id = step_run
        elif isinstance(step_run, dict):
            step_run_id = step_run.get("@id")
        return step_run_id

    def get_encode_file_matching_md5_of_blob(self, file):
        """Finds an ENCODE File object whose md5sum matches md5 of a blob in URI in backend.

        Args:
            file (str): String representing an URI to an object in the backend.

        Returns:
            dict: Dictionary representation of the matching file object on portal.
            None if no matching objects are found.
        """
        md5sum = self.backend.md5sum(file)
        search_param = [("md5sum", md5sum), ("type", "File")]
        encode_files = self.conn.search(search_param)
        filtered_encode_files = type(self).filter_encode_files_by_status(encode_files)
        if filtered_encode_files:
            if len(filtered_encode_files) > 1:
                self.logger.warning(
                    "get_encode_file_matching_md5_of_blob found more than 1 files matching the md5 of the blob."
                )
            return self.conn.get(filtered_encode_files[0].get("@id"))
        else:
            return None

    @staticmethod
    def filter_encode_files_by_status(
        encode_files, forbidden_statuses=("replaced", "revoked", "deleted")
    ):
        """Filter out files whose statuses are not allowed.

        From list of dicts representing encode file objects, filter out ones whose statuses are not allowed.

        Args:
            encode_files (list): List containing dicts representing encode file objects.
            forbidden_statuses (list): List of statuses. If file object has one of these statuses it will be filtered out.
            Statuses that encode file can have are: uploading, upload failed, in progress, released, archived, deleted, replaced,
            revoked, content error.

        Returns:
            list: List containing the dicts whose statuses are not contained in forbidden_statuses, empty list is possible.

        Raises:
            KeyError: If some of the files do not have status (this is an indication of an error on portal).
        """
        filtered_files = [
            file for file in encode_files if file["status"] not in forbidden_statuses
        ]
        return filtered_files

    def raw_files_accessioned(self):
        for file in self.analysis.raw_fastqs:
            if not self.get_encode_file_matching_md5_of_blob(file.filename):
                return False
        return True

    def accession_file(self, encode_file, gs_file):
        file_exists = self.get_encode_file_matching_md5_of_blob(gs_file.filename)
        submitted_file_path = {"submitted_file_name": gs_file.filename}
        if file_exists:
            self.logger.warning(
                "%s Attempting to post duplicate file of %s with md5sum %s",
                type(self).ACCESSION_LOG_KEY,
                file_exists.get("accession"),
                encode_file.get("md5sum"),
            )
        local_file = self.backend.download(gs_file.filename)[0]
        encode_file["submitted_file_name"] = local_file
        encode_posted_file = self.conn.post(encode_file)
        os.remove(local_file)
        encode_posted_file = self.patch_file(encode_posted_file, submitted_file_path)
        self.new_files.append(encode_posted_file)
        return encode_posted_file
        return file_exists

    def patch_file(self, encode_file, new_properties):
        new_properties[self.conn.ENCID_KEY] = encode_file.get("accession")
        return self.conn.patch(new_properties, extend_array_values=False)

    def log_if_exists(self, payload, profile_key):
        """
        If an object with given aliases already exists, as determined by an additional GET request,
        then log a warning before attempting to POST the payload. Truthiness of the dict returned by
        encode_utils.connection.Connection.get() is sufficient to check the object's existence on
        the portal, since it returns an empty dict when no matching record is found.
        """

        aliases = payload.get("aliases")
        if aliases:
            if self.conn.get(aliases, database=True):
                self.logger.error(
                    "%s %s with aliases %s already exists, will not post it",
                    profile_key.capitalize().replace("_", " "),
                    type(self).ACCESSION_LOG_KEY,
                    aliases,
                )

    def get_or_make_step_run(self, lab_prefix, run_name, step_version, task_name):
        """
        encode_utils.connection.Connection.post() does not fail on alias conflict, and does not
        expose the response status code, so we need to check for the existence of the object first
        before attempting to POST it with Accession.log_if_exists().
        """
        docker_tag = self.analysis.get_tasks(task_name)[0].docker_image.split(":")[1]
        aliases = [
            "{}:{}-{}-{}".format(
                lab_prefix, run_name, self.analysis.workflow_id, docker_tag
            )
        ]
        payload = {
            "aliases": aliases,
            "status": "in progress",
            "analysis_step_version": step_version,
        }
        profile_key = "analysis_step_runs"
        self.log_if_exists(payload, profile_key)
        payload[type(self).PROFILE_KEY] = profile_key
        return self.conn.post(payload)

    @property
    @abstractmethod
    def assembly(self):
        """
        A reminder that subclasses of Accession *must* provide their own implementation
        for assembly.
        """
        raise NotImplementedError(
            (
                "This method should be implemented by concrete derived classes specific to"
                " the pipeline in question."
            )
        )

    @property
    def genome_annotation(self):
        """
        Not every pipeline will strictly need this method, so the @abstractmethod
        decorator is not required as in the case of assembly, but we still need a
        default implementation to so that file_from_template can check if the annotation
        is there.
        """
        return None

    def find_portal_property_from_filekey(
        self, filekey: str, portal_property: str
    ) -> str:
        """
        Generic helper method that all pipelines can use to find the annotation
        """
        files = self.analysis.get_files(filekey=filekey)
        if files:
            annotation = self.get_encode_file_matching_md5_of_blob(files[0].filename)
            return annotation.get(portal_property, "")
        else:
            raise KeyError(
                "Could not find any file with key {} in metadata".format(filekey)
            )

    @property
    def lab_pi(self):
        return self.COMMON_METADATA["lab"].split("/labs/")[1].split("/")[0]

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.get_encode_file_matching_md5_of_blob(
                self.analysis.raw_fastqs[0].filename
            ).get("dataset")
            return self._dataset
        else:
            return self._dataset

    @property
    def assay_term_name(self):
        return self.conn.get(self.dataset).get("assay_term_name")

    def get_number_of_biological_replicates(self):
        bio_reps = set(
            [
                rep.get("biological_replicate_number")
                for rep in self.conn.get(self.dataset).get("replicates")
            ]
        )
        return len(bio_reps)

    @property
    def is_replicated(self):
        return True if self.get_number_of_biological_replicates() > 1 else False

    def file_from_template(
        self,
        file,
        file_format,
        output_type,
        step_run,
        derived_from,
        dataset,
        file_format_type=None,
        extras: Optional[Dict[str, Any]] = None,
    ):
        """
        Note that extras are used to update the dictionary at the very end. This allows
        for callbacks to override properties like output_type that may have been
        specified in the template.
        """
        file_name = file.filename.split("gs://")[-1].replace("/", "-")
        obj = {
            "status": "uploading",
            "aliases": ["{}:{}".format(self.lab_pi, file_name)],
            "file_format": file_format,
            "output_type": output_type,
            "assembly": self.assembly,
            "dataset": dataset,
            "step_run": step_run.get("@id"),
            "derived_from": derived_from,
            "file_size": file.size,
            "md5sum": file.md5sum,
        }
        if file_format_type:
            obj["file_format_type"] = file_format_type
        if self.genome_annotation:
            obj["genome_annotation"] = self.genome_annotation
        if extras is not None:
            obj.update(extras)
        obj[type(self).PROFILE_KEY] = "file"
        obj.update(self.COMMON_METADATA)
        return obj

    def get_derived_from_all(self, file, files):
        ancestors = []
        for ancestor in files:
            ancestors.append(
                self.get_derived_from(
                    file,
                    ancestor.get("derived_from_task"),
                    ancestor.get("derived_from_filekey"),
                    ancestor.get("derived_from_output_type"),
                    ancestor.get("derived_from_inputs"),
                    ancestor.get("allow_empty"),
                )
            )
        return list(self.flatten(ancestors))

    def flatten(self, nested_list):
        if isinstance(nested_list, str):
            yield nested_list
        if isinstance(nested_list, list):
            for item in nested_list:
                yield from self.flatten(item)

    # Returns list of accession ids of files on portal or recently accessioned
    def get_derived_from(
        self,
        file,
        task_name,
        filekey,
        output_type=None,
        inputs=False,
        allow_empty=False,
    ):
        derived_from_files = self.analysis.search_up(
            file.task, task_name, filekey, inputs
        )
        encode_files = [
            self.get_encode_file_matching_md5_of_blob(gs_file.filename)
            for gs_file in derived_from_files
        ]
        accessioned_files = encode_files + self.new_files
        accessioned_files = [x for x in accessioned_files if x is not None]
        derived_from_accession_ids = []
        for gs_file in derived_from_files:
            for encode_file in accessioned_files:
                if gs_file.md5sum == encode_file.get("md5sum"):
                    # Optimal peaks can be mistaken for conservative peaks
                    # when their md5sum is the same
                    if output_type and output_type != encode_file.get("output_type"):
                        continue
                    derived_from_accession_ids.append(encode_file["@id"])
        derived_from_accession_ids = list(set(derived_from_accession_ids))

        # Raise exception when some or all of the derived_from files
        # are missing from the portal

        missing = "\n".join(
            [
                "{}: {}".format(filekey, filename)
                for filename in map(lambda x: x.filename, derived_from_files)
            ]
        )
        if not derived_from_accession_ids and not allow_empty:
            raise Exception(
                f"Missing all of the derived_from files on the portal: {missing}"
            )
        if len(derived_from_accession_ids) != len(derived_from_files):
            raise Exception(
                f"Missing some of the derived_from files on the portal: {missing}"
            )
        return derived_from_accession_ids

    # File object to be accessioned
    def make_file_obj(
        self,
        file,
        file_format,
        output_type,
        step_run,
        derived_from_files,
        file_format_type=None,
        callbacks: Optional[List[str]] = None,
    ):
        derived_from = self.get_derived_from_all(file, derived_from_files)
        extras: Dict[str, Any] = {}
        if callbacks:
            for callback in callbacks:
                result: Dict[str, Any] = getattr(self, callback)(file)
                extras.update(result)
        return self.file_from_template(
            file,
            file_format,
            output_type,
            step_run,
            derived_from,
            self.dataset,
            file_format_type,
            extras=extras,
        )

    def get_bio_replicate(self, encode_file, string=True):
        replicate = encode_file.get("biological_replicates")[0]
        if string:
            return str(replicate)
        return int(replicate)

    def post_qcs(self):
        for qc in self.raw_qcs:
            qc.payload.update({"quality_metric_of": qc.files})
            self.new_qcs.append(self.conn.post(qc.payload, require_aliases=False))

    def queue_qc(self, qc, encode_file, profile, shared=False):
        step_run_id = self.get_step_run_id(encode_file)
        qc.update({"step_run": step_run_id, "status": "in progress"})
        if self.assay_term_name:
            qc["assay_term_name"] = self.assay_term_name
        qc.update(self.COMMON_METADATA)
        qc[type(self).PROFILE_KEY] = profile
        # Shared QCs will have two or more file ids
        # under the 'quality_metric_of' property
        # and payload must be the same for all
        if shared:
            for item in self.raw_qcs:
                if item.payload == qc:
                    item.files.append(encode_file.get("@id"))
                    return
        self.raw_qcs.append(QualityMetric(qc, encode_file.get("@id")))

    def file_has_qc(self, encode_file, qc_name):
        if list(
            filter(lambda x: qc_name in x["@type"], encode_file["quality_metrics"])
        ):
            return True
        return False

    def get_attachment(self, gs_file, mime_type, add_ext=""):
        """
        Files with certain extensions will fail portal validation since it can't guess
        the mime type correctly, e.g. a `.log` file with mime type `text/plain` will
        cause a schema validation error. We can trick the portal by appending a dummy
        extension that will cause the portal to correctly guess the mime type, for
        instance in the above case appending a `.txt` extension will validate properly.
        """
        contents = self.backend.read_file(gs_file.filename)
        contents = b64encode(contents)
        if type(contents) is bytes:
            # The Portal treats the contents as string "b'bytes'"
            contents = str(contents).replace("b", "", 1).replace("'", "")
        obj = {
            "type": mime_type,
            "download": gs_file.filename.split("/")[-1] + add_ext,
            "href": "data:{};base64,{}".format(mime_type, contents),
        }
        return obj

    def accession_step(self, single_step_params):
        """
        Note that this method will attempt a getattr() when converting the qc method defined in the
        accessioning template to a function name. This will raise a NotImplementedError if the
        method is not defined, wrapping the AttributeError raised by getattr(). Quality metric
        helper functions should be implemented by derived classes.
        The optional parameter "requries_replication" is used to denote wdl tasks that
        will not be present in the metadata if the pipeline is ran on unreplicated data,
        for example pooled IDR in the ChIP-seq pipeline.
        """
        if single_step_params.get("requires_replication") is True:
            if not self.is_replicated:
                return
        step_run = self.get_or_make_step_run(
            self.lab_pi,
            single_step_params["dcc_step_run"],
            single_step_params["dcc_step_version"],
            single_step_params["wdl_task_name"],
        )
        accessioned_files = []
        for task in self.analysis.get_tasks(single_step_params["wdl_task_name"]):
            for file_params in single_step_params["wdl_files"]:
                for wdl_file in [
                    file
                    for file in task.output_files
                    if file_params["filekey"] in file.filekeys
                ]:

                    # Conservative IDR thresholded peaks may have
                    # the same md5sum as optimal one
                    try:
                        obj = self.make_file_obj(
                            wdl_file,
                            file_params["file_format"],
                            file_params["output_type"],
                            step_run,
                            file_params["derived_from_files"],
                            file_format_type=file_params.get("file_format_type"),
                            callbacks=file_params.get("callbacks"),
                        )
                        encode_file = self.accession_file(obj, wdl_file)
                    except Exception as e:
                        if "Conflict" in str(e) and file_params.get(
                            "possible_duplicate"
                        ):
                            continue
                        elif "Missing" in str(e):
                            raise
                        else:
                            self.logger.critical(
                                "An error occurred accessioning a file", exc_info=True
                            )
                            raise

                    # Parameter file inputted assumes Accession implements
                    # the methods to attach the quality metrics
                    quality_metrics = file_params.get("quality_metrics", [])
                    for qc in quality_metrics:
                        qc_method = getattr(self, type(self).QC_MAP[qc])
                        # Pass encode file with
                        # calculated properties
                        qc_method(
                            self.conn.get(encode_file.get("accession"), database=True),
                            wdl_file,
                        )
                    accessioned_files.append(encode_file)
        return accessioned_files

    def accession_steps(self):
        for step in self.steps.content:
            self.accession_step(step)
        self.post_qcs()


class AccessionGenericRna(Accession):
    def make_generic_correlation_qc(self, encode_file, gs_file, handler):
        """
        Make correlation QC metrics in  a pipeline agnostic fashion. Pipeline specific logic is
        taken care of in the handler, the function that formats the qc metric dictionary.
        """
        if (
            self.file_has_qc(encode_file, "CorrelationQualityMetric")
            or self.get_number_of_biological_replicates() != 2
        ):
            return
        qc = handler(gs_file)
        return self.queue_qc(qc, encode_file, "correlation-quality-metric", shared=True)


class AccessionLongReadRna(AccessionGenericRna):
    QC_MAP = {
        "long_read_rna_mapping": "make_long_read_rna_mapping_qc",
        "long_read_rna_quantification": "make_long_read_rna_quantification_qc",
        "long_read_rna_correlation": "make_long_read_rna_correlation_qc",
    }

    @property
    def assembly(self):
        filekey = "annotation_gtf"
        return self.find_portal_property_from_filekey(filekey, ASSEMBLY)

    @property
    def genome_annotation(self):
        filekey = "annotation_gtf"
        return self.find_portal_property_from_filekey(filekey, GENOME_ANNOTATION)

    def make_long_read_rna_correlation_qc(self, encode_file, gs_file):
        """
        Make and post Spearman QC for long read RNA by giving the make_generic_correlation_qc the
        appropriate handler.
        """
        return self.make_generic_correlation_qc(
            encode_file, gs_file, handler=self.prepare_long_read_rna_correlation_qc
        )

    def prepare_long_read_rna_correlation_qc(self, gs_file):
        """
        Handler for creating the correlation QC object, specifically for long read rna. Finds and
        parses the spearman QC JSON.
        """
        qc_file, *_ = self.analysis.search_down(
            gs_file.task, "calculate_spearman", "spearman"
        )
        qc = self.backend.read_json(qc_file)
        spearman_value = qc["replicates_correlation"]["spearman_correlation"]
        spearman_correlation_qc = {"Spearman correlation": spearman_value}
        return spearman_correlation_qc

    def make_long_read_rna_mapping_qc(self, encode_file, gs_file):
        """
        The commented lines add number_of_mapped_reads to the qc object, a field that is currently
        not valid under the schema.
        """
        if self.file_has_qc(encode_file, "LongReadRnaMappingQualityMetric"):
            return
        qc_file = self.analysis.get_files(filename=gs_file.task.outputs["mapping_qc"])[
            0
        ]
        qc = self.backend.read_json(qc_file)
        output_qc = {}
        mr = "mapping_rate"
        flnc = qc["full_length_non_chimeric_reads"]["flnc"]
        output_qc["full_length_non_chimeric_read_count"] = int(flnc)
        output_qc[mr] = float(qc[mr][mr])
        return self.queue_qc(
            output_qc, encode_file, "long-read-rna-mapping-quality-metric"
        )

    def make_long_read_rna_quantification_qc(self, encode_file, gs_file):
        if self.file_has_qc(encode_file, "LongReadRnaQuantificationQualityMetric"):
            return
        ngd = "number_of_genes_detected"
        qc_file = self.analysis.get_files(filename=gs_file.task.outputs[ngd])[0]
        qc = self.backend.read_json(qc_file)
        output_qc = {"genes_detected": int(qc[ngd][ngd])}
        return self.queue_qc(
            output_qc, encode_file, "long-read-rna-quantification-quality-metric"
        )


class AccessionMicroRna(AccessionGenericRna):
    QC_MAP = {
        "mirna_mapping": "make_microrna_mapping_qc",
        "mirna_quantification": "make_microrna_quantification_qc",
        "mirna_correlation": "make_microrna_correlation_qc",
        "star": "make_star_qc_metric",
    }

    @property
    def assembly(self):
        filekey = "annotation"
        return self.find_portal_property_from_filekey(filekey, ASSEMBLY)

    @property
    def genome_annotation(self):
        filekey = "annotation"
        return self.find_portal_property_from_filekey(filekey, GENOME_ANNOTATION)

    def make_microrna_quantification_qc(self, encode_file, gs_file):
        if self.file_has_qc(encode_file, "MicroRnaQuantificationQualityMetric"):
            return
        qc_file = self.analysis.get_files(
            filename=gs_file.task.outputs["star_qc_json"]
        )[0]
        qc = self.backend.read_json(qc_file)
        expressed_mirnas_qc = qc["expressed_mirnas"]
        return self.queue_qc(
            expressed_mirnas_qc, encode_file, "micro-rna-quantification-quality-metric"
        )

    def make_microrna_mapping_qc(self, encode_file, gs_file):
        if self.file_has_qc(encode_file, "MicroRnaMappingQualityMetric"):
            return
        qc_file = self.analysis.get_files(
            filename=gs_file.task.outputs["star_qc_json"]
        )[0]
        qc = self.backend.read_json(qc_file)
        aligned_reads_qc = qc["aligned_reads"]
        return self.queue_qc(
            aligned_reads_qc, encode_file, "micro-rna-mapping-quality-metric"
        )

    def make_microrna_correlation_qc(self, encode_file, gs_file):
        """
        Returns without queueing this QC for posting if the experiment is not replicated, since
        correlation is computed between pairs of replicates.
        """
        if (
            self.file_has_qc(encode_file, "CorrelationQualityMetric")
            or self.get_number_of_biological_replicates() != 2
        ):
            return
        qc_file = self.analysis.search_down(
            gs_file.task, "spearman_correlation", "spearman_json"
        )[0]
        qc = self.backend.read_json(qc_file)
        spearman_value = qc["spearman_correlation"]["spearman_correlation"]
        spearman_correlation_qc = {"Spearman correlation": spearman_value}
        return self.queue_qc(
            spearman_correlation_qc,
            encode_file,
            "correlation-quality-metric",
            shared=True,
        )

    def make_star_qc_metric(self, encode_bam_file, gs_file):
        if self.file_has_qc(encode_bam_file, "StarQualityMetric"):
            return
        qc_file = self.analysis.get_files(
            filename=gs_file.task.outputs["star_qc_json"]
        )[0]
        qc = self.backend.read_json(qc_file)
        star_qc_metric = qc.get("star_qc_metric")
        del star_qc_metric["Started job on"]
        del star_qc_metric["Started mapping on"]
        del star_qc_metric["Finished on"]
        for key, value in star_qc_metric.items():
            star_qc_metric[key] = string_to_number(value)
        return self.queue_qc(star_qc_metric, encode_bam_file, "star-quality-metric")


class AccessionChip(Accession):
    QC_MAP = {
        "chip_alignment": "make_chip_alignment_qc",
        "chip_align_enrich": "make_chip_align_enrich_qc",
        "chip_library": "make_chip_library_qc",
        "chip_replication": "make_chip_replication_qc",
        "chip_peak_enrichment": "make_chip_peak_enrichment_qc",
    }

    @property
    def assembly(self):
        files = self.analysis.get_files(filekey="ref_fa")
        if files:
            portal_index = self.get_encode_file_matching_md5_of_blob(files[0].filename)
        return portal_index["assembly"]

    @staticmethod
    def get_chip_pipeline_replication_method(qc: Dict[str, Any]) -> str:
        """
        Checks the qc report for the pipeline type and returns the appropriate
        reproducibility criteria, `idr` when using SPP peak caller and `overlap` if the
        peak caller was MACS2.
        """
        peak_caller = qc["general"]["peak_caller"]
        if peak_caller == "macs2":
            return "overlap"
        return "idr"

    def get_chip_pipeline_replicate(self, gs_file):
        """
        Searches for the input fastq array corresponding to the ancestor input fastqs of the current
        file and returns the pipeline replicate number. We only need to check R1, since it will
        always be there in both the single and paired ended runs of the ChIP pipeline. We need this
        in order to be able to identify the correct QC in the QC JSON.
        """
        parent_fastqs = [
            file.filename
            for file in self.analysis.search_up(
                gs_file.task, "align", "fastqs_R1", inputs="true"
            )
        ]
        pipeline_rep = None
        for k, v in self.analysis.metadata["inputs"].items():
            if "fastqs" in k and "ctl" not in k:
                if sorted(v) == sorted(parent_fastqs):
                    pipeline_rep = k.split("_")[1]
                    break
        if not pipeline_rep:
            raise ValueError(
                "Could not determine pipeline replicate number for file {}".format(
                    gs_file
                )
            )
        return pipeline_rep

    def maybe_preferred_default(self, gs_file: GSFile) -> Dict[str, bool]:
        """
        For replicated ChIP-seq experiment, the exact file that is to be labeled with
        preferred_default=true may vary. As such, this callback is registered for any
        file that might need to have this value set in the steps JSON, and called at
        file object generation time (make_file_obj) to fill in (or not) the missing
        value.
        """
        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])[
            "replication"
        ]["reproducibility"]["idr"]

        optimal_set = qc["opt_set"]
        current_set = gs_file.task.inputs["prefix"]
        if current_set == optimal_set:
            return {"preferred_default": True}
        return {}

    def maybe_conservative_set(self, gs_file: GSFile) -> Dict[str, str]:
        """
        For replicated ChIP-seq experiment, the exact file that is to be labeled as
        the conservative set may vary. As such, this callback is registered for any
        file that might need to have this value set in the steps JSON, and called at
        file object generation time (make_file_obj) to fill in (or not) the missing
        value.
        """
        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])[
            "replication"
        ]["reproducibility"]["idr"]

        consv_set = qc["consv_set"]
        current_set = gs_file.task.inputs["prefix"]
        if current_set == consv_set:
            return {"output_type": "conservative IDR thresholded peaks"}
        return {}

    def add_mapped_read_length(self, gs_file: GSFile) -> Dict[str, int]:
        """
        Obtains the value of mapped_read_length to post for bam files from the read
        length log in the ancestor align task in the ChIP-seq pipeline.
        """
        read_len_log = self.analysis.search_up(gs_file.task, "align", "read_len_log")[0]
        log_contents = self.backend.read_file(read_len_log.filename)
        try:
            mapped_read_length = int(log_contents)
        except ValueError as e:
            raise RuntimeError(
                f"Could not parse read length log into integer: tried to parse {log_contents}"
            ) from e
        return {"mapped_read_length": mapped_read_length}

    def make_chip_alignment_qc(self, encode_file, gs_file):
        """
        This function typecasts to match the ENCODE schema. Trucated zero values could
        potentially be deserialized from the qc json as integers instead of floats.
        """
        if self.file_has_qc(encode_file, "ChipAlignmentQualityMetric"):
            return
        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
        replicate = self.get_chip_pipeline_replicate(gs_file)
        file_output_type = encode_file.get("output_type")
        if "unfiltered" in file_output_type:
            qc_key, processing_stage = "samstat", "unfiltered"
        else:
            qc_key, processing_stage = "nodup_samstat", "filtered"
        output_qc = qc["align"][qc_key][replicate]
        for k, v in output_qc.items():
            if k.startswith("pct"):
                output_qc[k] = float(v)
            else:
                output_qc[k] = int(v)
        # Add after to avoid trying to cast
        output_qc["processing_stage"] = processing_stage
        return self.queue_qc(
            output_qc, encode_file, "chip-alignment-samstat-quality-metric"
        )

    def make_chip_align_enrich_qc(self, encode_file, gs_file):
        """
        The xcor plots are not downstream of encode_file, in fact, they don't even share
        a common parent task with encode_file. Instead, we search up to find the parent
        align task of the current filtered bam, find the corresponding align_R1 task
        with the same fastq input, and search downstream from there for the xcor plot.
        """
        if self.file_has_qc(encode_file, "ChipAlignmentEnrichmentQualityMetric"):
            return
        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
        replicate = self.get_chip_pipeline_replicate(gs_file)
        key_to_match = "fastqs_R1"
        parent_fastqs = [
            file.filename
            for file in self.analysis.search_up(
                gs_file.task, "align", key_to_match, inputs="true"
            )
        ]
        align_r1_tasks = self.analysis.get_tasks("align_R1")
        start_task = [
            i for i in align_r1_tasks if i.inputs[key_to_match] == parent_fastqs
        ]
        if len(start_task) != 1:
            raise ValueError(
                (
                    f"Incorrect number of candidate start tasks with {key_to_match}: "
                    f"expected 1 but found {start_task}"
                )
            )
        cross_corr_plot_pdf = self.analysis.search_down(
            start_task[0], "xcor", "plot_pdf"
        )[0]
        fingerprint_plot_png = self.analysis.search_down(gs_file.task, "jsd", "plot")[0]
        gc_bias_plot_png = self.analysis.search_down(
            gs_file.task, "gc_bias", "gc_plot"
        )[0]
        output_qc = {
            **qc["align_enrich"]["xcor_score"][replicate],
            **qc["align_enrich"]["jsd"][replicate],
        }
        # Typecasting to match ENCODE schema
        for k, v in output_qc.items():
            if k in [
                "argmin_corr",
                "estimated_fragment_len",
                "phantom_peak",
                "subsampled_reads",
            ]:
                output_qc[k] = int(v)
            else:
                output_qc[k] = float(v)
        output_qc.update(
            {
                "cross_correlation_plot": self.get_attachment(
                    cross_corr_plot_pdf, "application/pdf"
                ),
                "jsd_plot": self.get_attachment(fingerprint_plot_png, "image/png"),
                "gc_bias_plot": self.get_attachment(gc_bias_plot_png, "image/png"),
            }
        )
        return self.queue_qc(
            output_qc, encode_file, "chip-alignment-enrichment-quality-metric"
        )

    def make_chip_library_qc(self, encode_file, gs_file):
        if self.file_has_qc(encode_file, "ChipLibraryQualityMetric"):
            return
        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
        replicate = self.get_chip_pipeline_replicate(gs_file)
        output_qc = {
            **qc["align"]["dup"][replicate],
            **qc["lib_complexity"]["lib_complexity"][replicate],
        }
        # Typecasting to match ENCODE schema
        for k, v in output_qc.items():
            if k in ["NRF", "PBC1", "PBC2", "pct_duplicate_reads"]:
                output_qc[k] = float(v)
            else:
                output_qc[k] = int(v)
        return self.queue_qc(output_qc, encode_file, "chip-library-quality-metric")

    def make_chip_replication_qc(self, encode_file, gs_file):
        """
        Rescue ratio and self-consistency ratio are only reported for optimal set. This
        set is determined by checking the QC JSON, and comparing to the prefix in the
        IDR task input in the WDL.
        The value of the QC's `reproducible_peaks` depends on the replicates or
        psuedo-replicates being compared.
        IDR cutoff, plot, and log are always reported for all IDR thresholded peaks
        files. They are not reported for the histone pipeline, which uses overlap.
        The IDR log file attachment is fudged with a .txt extension so that the portal
        can guess the mime type correctly and accept the file as valid.
        """
        if self.file_has_qc(encode_file, "ChipReplicationQualityMetric"):
            return
        raw_qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
        method = self.get_chip_pipeline_replication_method(raw_qc)
        qc = raw_qc["replication"]["reproducibility"][method]

        optimal_set = qc["opt_set"]
        current_set = gs_file.task.inputs["prefix"]
        output_qc = {}

        if current_set == optimal_set:
            output_qc.update(
                {
                    k: v
                    for k, v in qc.items()
                    if k
                    in ["rescue_ratio", "self_consistency_ratio", "reproducibility"]
                }
            )

        task_name = gs_file.task.task_name
        num_peaks = None
        if task_name == f"{method}_ppr":
            num_peaks = qc["Np"]
        elif task_name in ["idr", "overlap"]:
            num_peaks = qc["Nt"]
        elif task_name == f"{method}_pr":
            rep_num = current_set.split("-")[0][-1]
            num_peaks = qc[f"N{rep_num}"]
        if num_peaks is not None:
            output_qc["reproducible_peaks"] = int(num_peaks)

        if method == "idr":
            output_qc["idr_cutoff"] = float(gs_file.task.inputs["idr_thresh"])
            idr_plot_png = self.analysis.get_files(
                filename=gs_file.task.outputs["idr_plot"]
            )[0]
            idr_log = self.analysis.get_files(filename=gs_file.task.outputs["idr_log"])[
                0
            ]
            output_qc.update(
                {"idr_dispersion_plot": self.get_attachment(idr_plot_png, "image/png")}
            )
            output_qc.update(
                {
                    "idr_parameters": self.get_attachment(
                        idr_log, "text/plain", add_ext=".txt"
                    )
                }
            )
        return self.queue_qc(output_qc, encode_file, "chip-replication-quality-metric")

    def make_chip_peak_enrichment_qc(self, encode_file, gs_file):
        """
        The peak region stats are only useful for the optimal set, since the ones for
        rep1 and rep2 are applicable to files that are not posted by to the portal.
        IDR frip scores are applicable to any pair undergoing IDR, so they are always
        looked for.
        """
        if self.file_has_qc(encode_file, "ChipPeakEnrichmentQualityMetric"):
            return

        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
        method = self.get_chip_pipeline_replication_method(qc)

        optimal_set = qc["replication"]["reproducibility"][method]["opt_set"]
        current_set = gs_file.task.inputs["prefix"]

        output_qc = {
            "frip": qc["peak_enrich"]["frac_reads_in_peaks"][method][current_set][
                "frip"
            ]
        }
        if current_set == optimal_set:
            output_qc.update({**qc["peak_stat"]["peak_region_size"][f"{method}_opt"]})
        for k, v in output_qc.items():
            if k in ["mean", "frip"]:
                output_qc[k] = float(v)
            else:
                output_qc[k] = int(v)
        return self.queue_qc(
            output_qc, encode_file, "chip-peak-enrichment-quality-metric"
        )


def accession_factory(
    pipeline_type: str, accession_metadata: str, *args, **kwargs
) -> Accession:
    """
    Matches against the user-specified pipeline_type string and returns an instance of
    the appropriate accession subclass. Usage of this factory has the nice effect of
    automatically supplying the appropriate AccessionSteps based on the pipeline name.
    """
    pipeline_type_map = {
        "mirna": AccessionMicroRna,
        "long_read_rna": AccessionLongReadRna,
        "chip_map_only": AccessionChip,
        "tf_chip_peak_call_only": AccessionChip,
        "histone_chip_peak_call_only": AccessionChip,
    }
    selected_accession: Optional[Type[Accession]] = None
    try:
        selected_accession = pipeline_type_map[pipeline_type]
    except KeyError as e:
        pipeline_type_options = ", ".join(pipeline_type_map.keys())
        raise RuntimeError(
            f"Could not find pipeline type {pipeline_type}: valid options are {pipeline_type_options}"
        ) from e
    current_dir = Path(__file__).resolve()
    steps_json_path = (
        current_dir.parents[1] / "accession_steps" / f"{pipeline_type}_steps.json"
    )
    accession_steps = AccessionSteps(steps_json_path)
    metadata = MetaData(accession_metadata)
    backend = kwargs.pop("backend", None)
    analysis = Analysis(
        metadata, raw_fastqs_keys=accession_steps.raw_fastqs_keys, backend=backend
    )
    return selected_accession(accession_steps, analysis, *args, **kwargs)



import json
from contextlib import contextmanager
from pathlib import Path
from types import GeneratorType
from typing import Dict, List
from unittest.mock import PropertyMock

import attr
import pytest
from pytest_mock.plugin import MockFixture
from requests import Response

from accession.accession import (
    Accession,
    AccessionLongReadRna,
    AccessionMicroRna,
    AccessionSteps,
    accession_factory,
)
from accession.analysis import Analysis, MetaData

from .fixtures import MockGCBackend, MockMetaData


@attr.s(auto_attribs=True)
class MockFile:
    filename: str
    size: int
    md5sum: str


@pytest.fixture
def mirna_accessioner(accessioner_factory):
    factory = accessioner_factory(
        metadata_file="mirna_replicated_metadata.json", assay_name="mirna"
    )
    accessioner, _ = next(factory)
    return accessioner


@pytest.mark.docker
@pytest.mark.filesystem
def test_get_encode_file_matching_md5_of_blob(mirna_accessioner):
    fastq = mirna_accessioner.analysis.raw_fastqs[0]
    portal_file = mirna_accessioner.get_encode_file_matching_md5_of_blob(fastq.filename)
    assert portal_file.get("fastq_signature")


@pytest.mark.docker
@pytest.mark.filesystem
def test_raw_files_accessioned(mirna_accessioner):
    assert mirna_accessioner.raw_files_accessioned()


@pytest.mark.docker
@pytest.mark.filesystem
def test_assembly(mirna_accessioner):
    assert mirna_accessioner.assembly == "mm10"


@pytest.mark.docker
@pytest.mark.filesystem
def test_genome_annotation(mirna_accessioner):
    assert mirna_accessioner.genome_annotation == "M21"


@pytest.fixture
def ok_response():
    r = Response()
    r.status_code = 200
    return r


class FakeConnection:
    def __init__(self, dcc_url, response=None):
        self._dcc_url = dcc_url
        self._response = response

    @property
    def dcc_url(self):
        return self._dcc_url

    def get(self, query):
        return self._response


def test_get_number_of_biological_replicates(
    ok_response, mock_metadata, mock_gc_backend
):
    x = AccessionSteps("path")
    analysis = Analysis(mock_metadata, backend=mock_gc_backend)
    connection = FakeConnection(
        "https://www.zencodeproject.borg",
        response={
            "replicates": [
                {"biological_replicate_number": 1},
                {"biological_replicate_number": 2},
            ]
        },
    )
    y = AccessionMicroRna(x, analysis, connection, "lab", "award")
    y._dataset = "my_dataset"
    assert y.get_number_of_biological_replicates() == 2


@pytest.mark.docker
@pytest.mark.skip(
    reason="Elasticsearch is not reliable, produces inconsistent results."
)
@pytest.mark.filesystem
def test_get_derived_from(mirna_accessioner):
    bowtie_step = mirna_accessioner.steps.content[0]
    analysis = mirna_accessioner.analysis
    task = analysis.get_tasks(bowtie_step["wdl_task_name"])[0]
    bam = [file for file in task.output_files if "bam" in file.filekeys][0]
    raw_fastq_inputs = list(set(mirna_accessioner.raw_fastq_inputs(bam)))
    accession_ids = [
        mirna_accessioner.get_encode_file_matching_md5_of_blob(file.filename).get(
            "accession"
        )
        for file in raw_fastq_inputs
    ]
    params = bowtie_step["wdl_files"][0]["derived_from_files"][0]
    ancestors = mirna_accessioner.get_derived_from(
        bam,
        params.get("derived_from_task"),
        params.get("derived_from_filekey"),
        params.get("derived_from_output_type"),
        params.get("derived_from_inputs"),
    )
    ancestor_accessions = [ancestor.split("/")[-2] for ancestor in ancestors]
    assert len(ancestor_accessions) == 3
    assert len(accession_ids) == len(ancestor_accessions)
    assert set(accession_ids) == set(ancestor_accessions)


@pytest.mark.docker
@pytest.mark.skip(
    reason="Elasticsearch is not reliable, produces inconsistent results."
)
@pytest.mark.filesystem
def test_get_derived_from_all(mirna_accessioner):
    bowtie_step = mirna_accessioner.steps.content[0]
    analysis = mirna_accessioner.analysis
    task = analysis.get_tasks(bowtie_step["wdl_task_name"])[0]
    bam = [file for file in task.output_files if "bam" in file.filekeys][0]
    raw_fastq_inputs = list(set(mirna_accessioner.raw_fastq_inputs(bam)))
    accession_ids = [
        mirna_accessioner.get_encode_file_matching_md5_of_blob(file.filename).get(
            "accession"
        )
        for file in raw_fastq_inputs
    ]
    derived_from_files = bowtie_step["wdl_files"][0]["derived_from_files"]
    ancestors = mirna_accessioner.get_derived_from_all(bam, derived_from_files)
    ancestor_accessions = [ancestor.split("/")[-2] for ancestor in ancestors]
    assert len(ancestor_accessions) == 3
    assert len(accession_ids) == len(ancestor_accessions)
    assert set(accession_ids) == set(ancestor_accessions)


def mock_post_step_run(payload):
    payload.update({"@type": ["AnalysisStepRun"]})
    return payload


@pytest.mark.docker
@pytest.mark.filesystem
def test_get_or_make_step_run(mocker, mirna_accessioner):
    mocker.patch.object(mirna_accessioner.conn, "post", mock_post_step_run)
    mocker.patch.object(mirna_accessioner, "log_if_exists", autospec=True)
    bowtie_step = mirna_accessioner.steps.content[0]
    step_run = mirna_accessioner.get_or_make_step_run(
        mirna_accessioner.lab_pi,
        bowtie_step["dcc_step_run"],
        bowtie_step["dcc_step_version"],
        bowtie_step["wdl_task_name"],
    )
    assert "AnalysisStepRun" in step_run.get("@type")
    step_version = step_run.get("analysis_step_version")
    assert bowtie_step["dcc_step_version"] == step_version


@pytest.mark.docker
@pytest.mark.filesystem
def test_accession_file(mirna_accessioner):
    bowtie_step = mirna_accessioner.steps.content[0]
    analysis = mirna_accessioner.analysis
    task = analysis.get_tasks(bowtie_step["wdl_task_name"])[0]
    bam = [file for file in task.output_files if "bam" in file.filekeys][0]
    step_run = mirna_accessioner.get_or_make_step_run(
        mirna_accessioner.lab_pi,
        bowtie_step["dcc_step_run"],
        bowtie_step["dcc_step_version"],
        bowtie_step["wdl_task_name"],
    )
    file_params = bowtie_step["wdl_files"][0]
    obj = mirna_accessioner.make_file_obj(
        bam,
        file_params["file_format"],
        file_params["output_type"],
        step_run,
        file_params["derived_from_files"],
        file_format_type=file_params.get("file_format_type"),
    )
    encode_file = mirna_accessioner.accession_file(obj, bam)
    assert encode_file.get("accession")
    assert encode_file.get("status") == "uploading"
    assert encode_file.get("submitted_file_name") == bam.filename


@pytest.mark.docker
@pytest.mark.filesystem
def test_make_file_obj(mirna_accessioner):
    bowtie_step = mirna_accessioner.steps.content[0]
    analysis = mirna_accessioner.analysis
    task = analysis.get_tasks(bowtie_step["wdl_task_name"])[0]
    bam = [file for file in task.output_files if "bam" in file.filekeys][0]
    step_run = mirna_accessioner.get_or_make_step_run(
        mirna_accessioner.lab_pi,
        bowtie_step["dcc_step_run"],
        bowtie_step["dcc_step_version"],
        bowtie_step["wdl_task_name"],
    )
    file_params = bowtie_step["wdl_files"][0]
    obj = mirna_accessioner.make_file_obj(
        bam,
        file_params["file_format"],
        file_params["output_type"],
        step_run,
        file_params["derived_from_files"],
        file_format_type=file_params.get("file_format_type"),
    )
    assert obj.get("md5sum") and obj.get("file_size")
    assert len(obj.get("derived_from")) == 3


def test_accession_init(mock_accession: Accession, lab: str, award: str) -> None:
    assert mock_accession.COMMON_METADATA["lab"] == lab
    assert mock_accession.COMMON_METADATA["award"] == award
    assert all(
        (
            mock_accession.analysis,
            mock_accession.backend,
            mock_accession.conn,
            mock_accession.steps,
        )
    )


def test_get_step_run_id_string(mock_accession: Accession) -> None:
    expected = "/analysis-step-runs/123foobar"
    encode_file = {"step_run": expected}
    result = mock_accession.get_step_run_id(encode_file)
    assert result == expected


def test_get_step_run_id_dict(mock_accession: Accession) -> None:
    expected = "/analysis-step-runs/123foobar"
    encode_file = {"step_run": {"@id": expected}}
    result = mock_accession.get_step_run_id(encode_file)
    assert result == expected


def test_lab_pi(mock_accession: Accession, lab: str, award: str) -> None:
    assert mock_accession.lab_pi == "encode-processing-pipeline"


@pytest.mark.parametrize("file_format_type", [None, "bedMethyl"])
def test_file_from_template(
    mocker: MockFixture,
    mock_accession: Accession,
    lab: str,
    award: str,
    mock_file: MockFile,
    file_format_type: str,
) -> None:
    kwargs = {
        "file": mock_file,
        "file_format": "tsv",
        "output_type": "gene quantifications",
        "step_run": {"@id": "my-step-run"},
        "derived_from": "baz",
        "dataset": "qux",
        "file_format_type": file_format_type,
    }
    file = mock_accession.file_from_template(**kwargs)  # type: ignore
    expected = {
        "status": "uploading",
        "aliases": ["encode-processing-pipeline:foo-bar"],
        "file_size": mock_file.size,
        "md5sum": mock_file.md5sum,
        "assembly": "hg19",
        "step_run": "my-step-run",
    }
    for k in ["output_type", "file_format", "dataset", "derived_from"]:
        expected[k] = kwargs[k]
    for k, v in expected.items():
        assert file[k] == v
    if file_format_type:
        assert file["file_format_type"] == file_format_type
    else:
        assert "file_format_type" not in file
    assert file["genome_annotation"] == "V19"
    assert "lab" in file
    assert "award" in file


def test_flatten(mock_accession: Accession) -> None:
    result = mock_accession.flatten([["a", "b"], ["c", "d"]])
    assert isinstance(result, GeneratorType)
    assert list(result) == ["a", "b", "c", "d"]


def test_get_bio_replicate_str(
    mock_accession: Accession, mock_encode_file: Dict[str, List[int]]
) -> None:
    result = mock_accession.get_bio_replicate(mock_encode_file)
    assert result == "1"


def test_get_bio_replicate_int(
    mock_accession: Accession, mock_encode_file: Dict[str, List[int]]
) -> None:
    result = mock_accession.get_bio_replicate(mock_encode_file, string=False)
    assert result == 1


def mock_queue_qc(qc, *args, **kwargs):
    return qc


@pytest.mark.filesystem
def test_make_microrna_quantification_qc(mock_replicated_mirna_accession):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_quantification_qc("foo", gs_file)
    assert qc == {"expressed_mirnas": 393}


@pytest.mark.filesystem
def test_make_microrna_mapping_qc(mock_replicated_mirna_accession):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_mapping_qc("foo", gs_file)
    assert qc == {"aligned_reads": 5873570}


@pytest.mark.filesystem
def test_make_microrna_correlation_qc_replicated(mock_replicated_mirna_accession):
    gs_file = [
        i for i in mock_replicated_mirna_accession.analysis.get_files(filekey="tsv")
    ][0]
    qc = mock_replicated_mirna_accession.make_microrna_correlation_qc("foo", gs_file)
    assert qc == {"Spearman correlation": 0.8885044458946942}


@pytest.mark.filesystem
def test_make_microrna_correlation_qc_unreplicated_returns_none(
    mocker, mock_accession_unreplicated, mirna_replicated_analysis
):
    mocker.patch.object(
        mock_accession_unreplicated, "analysis", mirna_replicated_analysis
    )
    mocker.patch.object(
        mock_accession_unreplicated, "backend", mirna_replicated_analysis.backend
    )
    mocker.patch.object(mock_accession_unreplicated, "file_has_qc", return_value=False)
    mocker.patch.object(mock_accession_unreplicated, "queue_qc", mock_queue_qc)
    mocker.patch.object(
        mock_accession_unreplicated,
        "get_number_of_biological_replicates",
        return_value=1,
    )
    gs_file = [
        i for i in mock_accession_unreplicated.analysis.get_files(filekey="tsv")
    ][0]
    qc = mock_accession_unreplicated.make_microrna_correlation_qc("foo", gs_file)
    assert qc is None


@pytest.mark.filesystem
def test_make_star_qc_metric(mock_replicated_mirna_accession):
    gs_file = [
        i
        for i in mock_replicated_mirna_accession.analysis.get_files(filekey="bam")
        if "rep1" in i.filename
    ][0]
    current_dir = Path(__file__).resolve()
    validation = current_dir.parent / "data" / "validation" / "mirna" / "files.json"
    with open(validation) as f:
        data = json.load(f)
    expected = [i for i in data if i["accession"] == "ENCFF590YAX"][0][
        "quality_metrics"
    ][0]
    qc = mock_replicated_mirna_accession.make_star_qc_metric("foo", gs_file)
    for k, v in qc.items():
        assert v == expected[k]


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize(
    "pipeline_type,condition,accessioner_class",
    [
        ("mirna", does_not_raise(), AccessionMicroRna),
        ("long_read_rna", does_not_raise(), AccessionLongReadRna),
        ("not_valid", pytest.raises(RuntimeError), None),
    ],
)
def test_accession_factory(
    mocker, mock_gc_backend, pipeline_type, condition, accessioner_class
):
    mocker.patch(
        "builtins.open",
        mocker.mock_open(read_data='{"workflowRoot": "gs://foo/bar", "calls": {}}'),
    )
    with condition:
        accessioner = accession_factory(
            pipeline_type, "metadata.json", "bar", "baz", "qux", backend=mock_gc_backend
        )
        assert isinstance(accessioner, accessioner_class)


@pytest.fixture
def mock_replicated_mirna_accession(mocker, mirna_replicated_analysis, mock_accession):
    """
    Contains a legitimate replicated mirna analysis for purposes of testing mirna qm
    generation without creating any connections to servers.
    """
    mocker.patch.object(mock_accession, "analysis", mirna_replicated_analysis)
    mocker.patch.object(mock_accession, "backend", mirna_replicated_analysis.backend)
    mocker.patch.object(mock_accession, "file_has_qc", return_value=False)
    mocker.patch.object(mock_accession, "queue_qc", mock_queue_qc)
    mocker.patch.object(
        mock_accession, "get_number_of_biological_replicates", return_value=2
    )
    return mock_accession


@pytest.fixture
def mirna_replicated_analysis(
    mock_accession_gc_backend: MockFixture
) -> Analysis:  # noqa: F811
    current_dir = Path(__file__).resolve()
    metadata_json_path = current_dir.parent / "data" / "mirna_replicated_metadata.json"
    analysis = Analysis(MetaData(metadata_json_path), backend=mock_accession_gc_backend)
    return analysis


@pytest.fixture
def mock_encode_file() -> Dict[str, List[int]]:
    return {"biological_replicates": [1, 2]}


@pytest.fixture
def mock_accession_unreplicated(
    mocker: MockFixture,
    mock_accession_gc_backend: MockGCBackend,
    mock_metadata: MockMetaData,
    lab: str,
    award: str,
) -> Accession:
    """
    Mocked accession instance with dummy __init__ that doesn't do anything and pre-baked
    assembly property. @properties must be patched before instantiation
    """
    mocker.patch.object(
        Accession, "is_replicated", new_callable=PropertyMock(return_value=False)
    )
    mocked_accession = AccessionMicroRna(
        "imaginary_steps.json",
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        "mock_server.biz",
        lab,
        award,
    )
    return mocked_accession


@pytest.fixture
def mock_file() -> MockFile:
    return MockFile("gs://foo/bar", 123, "abc")


def test_filter_encode_files_by_status_one_hit():
    files = [{"status": "foo"}, {"status": "bar"}, {"status": "deleted"}]
    assert len(Accession.filter_encode_files_by_status(files)) == 2


def test_filter_encode_files_by_status_no_hits():
    files = [{"status": "foo"}]
    assert len(Accession.filter_encode_files_by_status(files)) == 1



from contextlib import contextmanager

import pytest

from accession.accession import AccessionChip
from accession.file import GSFile
from accession.task import Task


@pytest.fixture
def gsfile():
    task = {
        "inputs": {
            "prefix": "pooled-pr1_vs_pooled-pr2",
            "fastqs_R1": ["gs://abc/spam.fastq.gz"],
        },
        "outputs": {},
    }
    my_task = Task("my_task", task, analysis="qux")
    return GSFile(
        "foo", "gs://abc/spam.fastq.gz", md5sum="123", size="456", task=my_task
    )


@contextmanager
def does_not_raise():
    """
    A dummy context manager useful for parametrized tests that may or may not raise
    an error. Usage of this expectation indicates that the current test parameter set
    should not raise an error.

    See http://doc.pytest.org/en/latest/example/parametrize.html#parametrizing-conditional-raising
    """
    yield


REP = "rep1"


@pytest.fixture
def mock_accession_patched_qc(mocker, mock_accession_chip, gsfile):
    """
    Performs shared patches to qc metric maker functions. Dependents on this fixture
    should patch mock_accession_chip.backend.read_json() to supply the appropriate qc stub
    """
    mocker.patch.object(mock_accession_chip, "file_has_qc", return_value=False)
    mocker.patch.object(
        mock_accession_chip.analysis, "get_files", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_chip, "get_chip_pipeline_replicate", return_value=REP
    )
    mocker.patch.object(mock_accession_chip, "queue_qc", lambda output, *args: output)
    return mock_accession_chip


"""
Ideally these would be pytest fixtures, but they don't play nicely when passed as
return_value in mocker.patch.object.
"""
align_qc_dup = {
    "align": {
        "samstat": {
            "rep1": {
                "total_reads": 31167664,
                "total_reads_qc_failed": 0,
                "duplicate_reads": 0,
                "duplicate_reads_qc_failed": 0,
                "mapped_reads": 28618875,
                "mapped_reads_qc_failed": 0,
                "pct_mapped_reads": 91.8,
                "paired_reads": 0,
                "paired_reads_qc_failed": 0,
                "read1": 0,
                "read1_qc_failed": 0,
                "read2": 0,
                "read2_qc_failed": 0,
                "properly_paired_reads": 0,
                "properly_paired_reads_qc_failed": 0,
                "pct_properly_paired_reads": 0.0,
                "with_itself": 0,
                "with_itself_qc_failed": 0,
                "singletons": 0,
                "singletons_qc_failed": 0,
                "pct_singletons": 0.0,
                "diff_chroms": 0,
                "diff_chroms_qc_failed": 0,
            }
        }
    }
}


align_qc_nodup = {
    "align": {
        "nodup_samstat": {
            "rep1": {
                "total_reads": 22033059,
                "total_reads_qc_failed": 0,
                "duplicate_reads": 0,
                "duplicate_reads_qc_failed": 0,
                "mapped_reads": 22033059,
                "mapped_reads_qc_failed": 0,
                "pct_mapped_reads": 100.0,
                "paired_reads": 0,
                "paired_reads_qc_failed": 0,
                "read1": 0,
                "read1_qc_failed": 0,
                "read2": 0,
                "read2_qc_failed": 0,
                "properly_paired_reads": 0,
                "properly_paired_reads_qc_failed": 0,
                "pct_properly_paired_reads": 0.0,
                "with_itself": 0,
                "with_itself_qc_failed": 0,
                "singletons": 0,
                "singletons_qc_failed": 0,
                "pct_singletons": 0.0,
                "diff_chroms": 0,
                "diff_chroms_qc_failed": 0,
            }
        }
    }
}


align_enrich_qc = {
    "align_enrich": {
        "xcor_score": {
            "rep1": {
                "subsampled_reads": 15000000,
                "estimated_fragment_len": 100,
                "corr_estimated_fragment_len": 0.215460537129363,
                "phantom_peak": 80,
                "corr_phantom_peak": 0.2136902,
                "argmin_corr": 1500,
                "min_corr": 0.1630407,
                "NSC": 1.321514,
                "RSC": 1.034953,
            }
        },
        "jsd": {
            "rep1": {
                "auc": 0.22069022540637373,
                "syn_auc": 0.49477458966709664,
                "x_intercept": 0.1425972317467426,
                "syn_x_intercept": 0.0,
                "elbow_pt": 0.6535464871014113,
                "syn_elbow_pt": 0.506542601727043,
                "jsd": 0.003923754877223693,
                "syn_jsd": 0.3758404209038387,
                "pct_genome_enrich": 28.538239921946694,
                "diff_enrich": 0.31028662370578,
                "ch_div": 0.002800808043644715,
            }
        },
    }
}


library_qc = {
    "align": {
        "dup": {
            "rep1": {
                "unpaired_reads": 0,
                "paired_reads": 31681025,
                "unmapped_reads": 0,
                "unpaired_duplicate_reads": 0,
                "paired_duplicate_reads": 1669558,
                "paired_optical_duplicate_reads": 32245,
                "pct_duplicate_reads": 5.269900000000001,
            }
        }
    },
    "lib_complexity": {
        "lib_complexity": {
            "rep1": {
                "total_fragments": 31665818,
                "distinct_fragments": 29998922,
                "positions_with_one_read": 1523995,
                "NRF": 0.94736,
                "PBC1": 0.946871,
                "PBC2": 18.638593,
            }
        }
    },
}


tf_replication_qc = {
    "general": {"pipeline_type": "tf", "peak_caller": "spp"},
    "replication": {
        "reproducibility": {
            "idr": {
                "Nt": 26692,
                "N1": 27036,
                "N2": 11788,
                "Np": 31098,
                "N_opt": 31098,
                "N_consv": 26692,
                "opt_set": "pooled-pr1_vs_pooled-pr2",
                "consv_set": "rep1_vs_rep2",
                "rescue_ratio": 1.1650681852240372,
                "self_consistency_ratio": 2.2935188327112317,
                "reproducibility": "borderline",
            }
        }
    },
}


histone_replication_qc = {
    "general": {"pipeline_type": "histone", "peak_caller": "macs2"},
    "replication": {
        "reproducibility": {
            "overlap": {
                **tf_replication_qc["replication"]["reproducibility"][  # type: ignore
                    "idr"
                ]  # type: ignore
            }
        }
    },
}


peak_enrichment_qc = {
    "general": {"pipeline_type": "tf", "peak_caller": "spp"},
    "replication": {
        "reproducibility": {"idr": {"opt_set": "pooled-pr1_vs_pooled-pr2"}}
    },
    "peak_stat": {
        "peak_region_size": {
            "idr_opt": {
                "min_size": 140.0,
                "25_pct": 385.0,
                "50_pct": 560.0,
                "75_pct": 560.0,
                "max_size": 2298.0,
                "mean": 482.61428387677665,
            }
        }
    },
    "peak_enrich": {
        "frac_reads_in_peaks": {
            "idr": {
                "rep1_vs_rep2": {"frip": 0.06048766636136033},
                "pooled-pr1_vs_pooled-pr2": {"frip": 0.06720758657484392},
            }
        }
    },
}


@pytest.mark.parametrize("peak_caller,expected", [("spp", "idr"), ("macs2", "overlap")])
def test_get_chip_pipeline_replication_method(peak_caller, expected):
    method = AccessionChip.get_chip_pipeline_replication_method(
        {"general": {"peak_caller": peak_caller}}
    )
    assert method == expected


@pytest.mark.parametrize(
    "qc,expected",
    [
        (
            b'{"replication": {"reproducibility": {"idr": {"opt_set": "pooled-pr1_vs_pooled-pr2"}}}}',
            {"preferred_default": True},
        ),
        (
            b'{"replication": {"reproducibility": {"idr": {"opt_set": "rep1_vs_rep2"}}}}',
            {},
        ),
    ],
)
def test_maybe_preferred_default(mocker, mock_accession_chip, gsfile, qc, expected):
    mocker.patch.object(
        mock_accession_chip.analysis, "get_files", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_chip.analysis.backend, "read_file", return_value=qc
    )
    result = mock_accession_chip.maybe_preferred_default(gsfile)
    assert result == expected


@pytest.mark.parametrize(
    "log_contents,expectation,expected_value",
    [(b"42", does_not_raise(), 42), (b"foo", pytest.raises(RuntimeError), 42)],
)
def test_add_chip_mapped_read_length(
    mocker, mock_accession_chip, gsfile, log_contents, expectation, expected_value
):
    mocker.patch.object(
        mock_accession_chip.analysis, "search_up", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_chip.analysis.backend, "read_file", return_value=log_contents
    )
    with expectation:
        result = mock_accession_chip.add_mapped_read_length(gsfile)
        assert result == {"mapped_read_length": expected_value}


@pytest.mark.parametrize(
    "processing_stage,align_qc",
    [("unfiltered", align_qc_dup), ("filtered", align_qc_nodup)],
)
def test_make_chip_alignment_qc(
    mocker, mock_accession_patched_qc, gsfile, processing_stage, align_qc
):
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_json", return_value=align_qc
    )
    output_type = "alignments"
    qc_key = "nodup_samstat"
    if processing_stage == "unfiltered":
        output_type = " ".join([processing_stage, output_type])
        qc_key = "samstat"
    encode_file = {"output_type": output_type}
    qc = mock_accession_patched_qc.make_chip_alignment_qc(encode_file, gs_file="foo")
    for k, v in qc.items():
        if k == "processing_stage":
            assert v == processing_stage
        else:
            assert v == align_qc["align"][qc_key][REP][k]


@pytest.fixture
def mock_accession_align_enrich(mocker, mock_accession_patched_qc, gsfile):
    mocker.patch.object(
        mock_accession_patched_qc.analysis, "search_up", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_patched_qc.analysis, "search_down", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_file", return_value=b"foo"
    )
    return mock_accession_patched_qc


def test_make_chip_align_enrich_qc(mocker, mock_accession_align_enrich, gsfile):
    mocker.patch.object(
        mock_accession_align_enrich.backend, "read_json", return_value=align_enrich_qc
    )
    task = {"inputs": {"fastqs_R1": ["gs://abc/spam.fastq.gz"]}, "outputs": {}}
    stub_task = Task("align_R1", task, analysis="qux")
    mocker.patch.object(
        mock_accession_align_enrich.analysis, "get_tasks", return_value=[stub_task]
    )
    encode_file = {}
    qc = mock_accession_align_enrich.make_chip_align_enrich_qc(
        encode_file, gs_file=gsfile
    )
    keys = ["xcor_score", "jsd"]
    for key in keys:
        for k, v in align_enrich_qc["align_enrich"][key][REP].items():
            assert qc[k] == v
    for key in ["cross_correlation_plot", "jsd_plot", "gc_bias_plot"]:
        for prop in ["type", "download", "href"]:
            assert prop in qc[key]


def test_make_chip_align_enrich_qc_raises(mocker, mock_accession_align_enrich, gsfile):
    mocker.patch.object(
        mock_accession_align_enrich.analysis, "get_tasks", return_value=[]
    )
    with pytest.raises(ValueError):
        mock_accession_align_enrich.make_chip_align_enrich_qc({}, gs_file=gsfile)


def test_make_chip_library_qc(mocker, mock_accession_patched_qc, gsfile):
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_json", return_value=library_qc
    )
    qc = mock_accession_patched_qc.make_chip_library_qc({}, gs_file="foo")
    for k, v in library_qc["align"]["dup"][REP].items():
        assert qc[k] == v
    for k, v in library_qc["lib_complexity"]["lib_complexity"][REP].items():
        assert qc[k] == v


def test_get_chip_pipeline_replicate(mocker, mock_accession_chip, gsfile):
    mocker.patch.object(
        mock_accession_chip.analysis, "search_up", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_chip.analysis,
        "metadata",
        {"inputs": {"fastqs_rep1_R1": ["gs://abc/spam.fastq.gz"]}},
    )
    rep = mock_accession_chip.get_chip_pipeline_replicate(gsfile)
    assert rep == "rep1"


def test_get_chip_pipeline_replicate_raises(mocker, mock_accession_chip, gsfile):
    mocker.patch.object(
        mock_accession_chip.analysis, "search_up", return_value=[gsfile]
    )
    mocker.patch.object(
        mock_accession_chip.analysis,
        "metadata",
        {"inputs": {"fastqs_rep1_R1": ["gs://abc/eggs.fastq.gz"]}},
    )
    with pytest.raises(ValueError):
        mock_accession_chip.get_chip_pipeline_replicate(gsfile)


@pytest.mark.parametrize(
    "task_name,current_set,reproducible_peaks_source",
    [
        ("idr_ppr", "pooled-pr1_vs_pooled-pr2", "Np"),
        ("idr", "rep1_vs_rep2", "Nt"),
        ("idr_pr", "rep1-pr1_vs_rep1-pr2", "N1"),
        ("overlap_ppr", "pooled-pr1_vs_pooled-pr2", "Np"),
        ("overlap", "rep1_vs_rep2", "Nt"),
        ("overlap_pr", "rep1-pr1_vs_rep1-pr2", "N1"),
    ],
)
def test_make_chip_replication_qc(
    mocker, mock_accession_patched_qc, task_name, current_set, reproducible_peaks_source
):
    """
    Tests 3 cases each for histone and TF pipelines. Histone QC shouldn't have IDR
    properties (plot, log, and cutoff). For both, only the optimal set should specify
    rescue ratio, self-consistency ratio, and reproducibility (pass/fail).
    """
    pipeline_qc = histone_replication_qc
    method = "overlap"
    if task_name.startswith("idr"):
        pipeline_qc = tf_replication_qc
        method = "idr"
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_json", return_value=pipeline_qc
    )
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_file", return_value=b"foo"
    )
    if method == "idr":
        idr_thresh = 0.05
        task = {
            "inputs": {"prefix": current_set, "idr_thresh": idr_thresh},
            "outputs": {"idr_log": "gs://abc/idr.log", "idr_plot": "gs://abc/plot.png"},
        }
    else:
        task = {"inputs": {"prefix": current_set}, "outputs": []}
    my_task = Task(task_name, task, analysis="qux")
    gsfile = GSFile(
        "foo", "gs://abc/spam.fastq.gz", md5sum="123", size="456", task=my_task
    )
    qc = mock_accession_patched_qc.make_chip_replication_qc({}, gs_file=gsfile)
    assert (
        qc["reproducible_peaks"]
        == pipeline_qc["replication"]["reproducibility"][method][
            reproducible_peaks_source
        ]
    )
    if method == "idr":
        assert qc["idr_cutoff"] == idr_thresh
        for k in ["idr_parameters", "idr_dispersion_plot"]:
            assert all(i in qc[k] for i in ["href", "type", "download"])
    optimal_keys = ["rescue_ratio", "self_consistency_ratio", "reproducibility"]
    if current_set == pipeline_qc["replication"]["reproducibility"][method]["opt_set"]:
        for k in optimal_keys:
            assert qc[k] == pipeline_qc["replication"]["reproducibility"][method][k]


@pytest.mark.parametrize(
    "current_set", [("pooled-pr1_vs_pooled-pr2"), ("rep1_vs_rep2")]
)
def test_make_chip_peak_enrichment_qc(mocker, mock_accession_patched_qc, current_set):
    """
    There is not a significant difference between the way this qc is generated between
    histone and tf pipelines. As such, only tf is tested here.
    """
    mocker.patch.object(
        mock_accession_patched_qc.backend, "read_json", return_value=peak_enrichment_qc
    )
    task = {"inputs": {"prefix": current_set}, "outputs": []}
    my_task = Task("idr", task, analysis="qux")
    gsfile = GSFile(
        "foo", "gs://abc/spam.fastq.gz", md5sum="123", size="456", task=my_task
    )
    qc = mock_accession_patched_qc.make_chip_peak_enrichment_qc({}, gs_file=gsfile)
    if (
        current_set
        == peak_enrichment_qc["replication"]["reproducibility"]["idr"]["opt_set"]
    ):
        for k, v in peak_enrichment_qc["peak_stat"]["peak_region_size"][
            "idr_opt"
        ].items():
            assert qc[k] == v
    assert (
        qc["frip"]
        == peak_enrichment_qc["peak_enrich"]["frac_reads_in_peaks"]["idr"][current_set][
            "frip"
        ]
    )
